#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

/*****************************************************************/
// !!! In order to avoid malfunction of the controller
// !!! and to ensure maximal protection for the tube
// !!! it is *essential* to activate the 
// !!! BROWN OUT DETECTION (BOD)!
/*****************************************************************/

// Connected to R2: drives T1
#define PORT_PWM PORTB  // OC0A   
#define DDR_PWM  DDRB   // OC0A
#define BIT_PWM  0      // OC0A

// Connected to R1/R4
#define PORT_SENSE  PORTB // ADC1 
#define DDR_SENSE   DDRB  // ADC1
#define BIT_SENSE   2     // ADC1

// Connected to R1/R4
// Hard wired to PORT_SENSE by PCB
#define PORT_SENSE2 PORTB // AIN1 
#define DDR_SENSE2  DDRB
#define BIT_SENSE2  1

// unused, connected to R3
#define PORT_SENSE3 PORTB // ADC3 
#define DDR_SENSE3  DDRB
#define BIT_SENSE3  3

#define IRQS_PER_SECOND        125
#define PRESCALE 8

static void ioinit(void); 
static void timers_init(void);

// Will be decremented every 8 milliseconds, if not already zero
volatile uint8_t count_8ms;

/*****************************************************************/

// Initialize Timer0 to make PWM  @ F_CPU/256 i.e. ~ 1kHz
// Initialize Timer1 to make IRQs @ 125kHz    i.e. every 8ms
void timers_init (void)
{
    // Timer1 in CTC-Mode and prescaled 8
    TCCR1 = (1 << CTC1) | (1 << CS12); 
    
    // PoutputCompare for desired Timer1 frequency
    OCR1C = (uint8_t) ((uint32_t) F_CPU / IRQS_PER_SECOND / PRESCALE -1);
    OCR1B = 0;
    
    // OutputCompare-Interrupt B for Timer 1
    TIMSK = (1 << OCIE1B);
    
    // PWM @ F_CPU/PWM_RESOLUTION at OC0A (PORT_PWM)
    TCCR0A = (1 << WGM00) | (1 << COM0A1);
    TCCR0B = (0 << CS01) | (1 << CS00); // Prescale = 1
}

/*****************************************************************/

// This ISR executes every 8 ms 
// and decrements `count_8ms' if it is not already zero
SIGNAL (SIG_OUTPUT_COMPARE1B)
{
    uint8_t val = count_8ms;
        
    if (val)
        count_8ms = val-1;
}    

/*****************************************************************/

// ADC enabled
// prescaled 2, i.e. runs @ 125kHz
#define ADC_VAL ((1 << ADEN) | (0 << ADPS2) | (0 << ADPS1) | (1 << ADPS0) | (1 << ADIF))

// Initialize hardware
void ioinit (void) 
{
    // high fuse = 0xcc (BOD = 4.3V, WDT=on)
    wdt_enable (WDTO_15MS);
    
    // all (unused) ports are IN + PullUp
    DDRB  = 0;
    PORTB = 0x3f;
    
    // V_REF = VCC
    // ADC1 (Port B2, Pad7)
    // right adjust result
    
    // ADC1 = B2 (Pad 7)
    // ADC2 = B4 (Pad 3)
    // ADC3 = B3 (Pad 2)
    
    ADMUX = (0 << REFS2) | (0 << REFS1) | (0 << ADLAR) | 1;
    
    // Enable Timer1 OutputCompare-IRQ
    // Set    Timer0 PWM
    timers_init();

    // SENSE = IN + highZ
    DDR_SENSE  &= ~(1 << BIT_SENSE);    
    PORT_SENSE &= ~(1 << BIT_SENSE);    
    
    // SENSE2 = IN + highZ (must be same as above)
    DDR_SENSE2  &= ~(1 << BIT_SENSE2);    
    PORT_SENSE2 &= ~(1 << BIT_SENSE2);    

    // SWITCH = OUT + LOW, T1 is high-active
    DDR_PWM  |=  (1 << BIT_PWM);    
    PORT_PWM &= ~(1 << BIT_PWM);    
}

/*****************************************************************/

int main (void)
{
    // ramp up PWM duty cycle 
    uint8_t startup = 1;
    
    // Run CPU @ 250kHz (8MHz / 32)
    CLKPR = (1 << CLKPCE);
    CLKPR = (1 << CLKPS2) | (1 << CLKPS0);

    // Initialize hardware
    ioinit (); 
    
    // IRQs are welcome
    sei();    

    // main loop
    while (0==0)
    {
        // arithmetic mean over 2**PP ADC values
        const uint8_t PP = 4;
        
        uint8_t i;
        uint16_t adc = 0;
        
        for (i=0; i < (1 << PP); i++)
        {
            // Start ADC single shot conversion
            ADCSRA = (1 << ADSC) | ADC_VAL;
            
            // Wait for ADC to complete
            while (!(ADCSRA & (1 << ADIF)))
                ;
                
            // add the ADC value to adc totals
            adc += ADC;    
        }

        // divide adc totals by 2**PP to get arithmetik mean
        adc >>= PP;

// The target's nominal voltage: this is why we are here...
#define U_SOLL 6.3   // V    

// These values describe the components on the PCB
#define R_HI  35.7   // kOhm // high side resistor R1
#define R_LO   9.9   // kOhm // low  side resistor R4
#define U_REF  5.0   // V    // U_REF is Vcc and comes from IC2, an LM2936Z-5
#define U_CE   0.75  // V    // emitter-collector dropout of transistor T1

#define ADC_RESOLUTION      1024.0
#define PWM_RESOLUTION       256.0

        const int32_t  K1 = PWM_RESOLUTION * ADC_RESOLUTION * U_SOLL;
        const int32_t  K2 = (R_HI+R_LO)/R_LO * U_REF;
        const int32_t  K3 = ADC_RESOLUTION * U_CE;
        const uint32_t K4 = PWM_RESOLUTION;

        // Compute value for OCR as 32-bit value from adc
        int32_t ocr_32;

        // Be careful with such computations,
        // do not shorten out `PWM_RESOLUTION' etc and mind the conditioning!
        ocr_32 = K1 / (K2 * adc - K3);
        ocr_32 = ocr_32 * ocr_32;
        ocr_32 /= K4;

        // Trunk 32-bit ocr value to an 8-bit saturated value
        uint8_t ocr_8 = ocr_32;
        
        if (ocr_32 >= 256)        ocr_8 = 255;
        if (ocr_32 < 0)           ocr_8 = 0;

        // Read actual OCR value    
        uint8_t ocr = OCR0A;

        // Soft start the voltage: rise PWM duty slowly,
        // i.e. increment it just every 4*8 ms.
        // In startup *and* countdown timer run dry?
        if (startup && 0 == count_8ms)
        {
            // Yes:
            if (ocr < ocr_8)
            {
                // OCR did not yet reach desired level of `ocr_8':
                // increment OCR and refuel the 8 ms countdown timer
                count_8ms = 4;
                ocr++;
            }
            else
                // OCR reached the desired level: startup is over
                startup = 0;
        }

        // After startup?
        if (!startup)
        {
            // Yes:
            // We react faster to a changing input voltage
            // Note: changing OCR is still slowed down by the ADC 
            if (ocr > ocr_8)        ocr--;
            if (ocr < ocr_8)        ocr++;
        }

        // Write the new OCR value
        // OCR will change by -1, 0 or 1
        OCR0A = ocr;    

        // Done
        wdt_reset();
    } // main loop
}
